# -*- coding: utf-8 -*-

import metrolopy as uc
import numpy as np

from metrolopy.tests.common import rand,make_gummy,make_number

def binary_func(f,df,sim=False,exp=None,fionly=False,uexp=-6,allowazero=True,
                allowbzero=True,allowlargeu=True):
    if sim:
        dof = dof=float('inf')
    else:
        dof = None
    a,ax,au,adof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                  uexp=uexp,allowzero=allowazero,
                                  allowlargeu=allowlargeu)
    if rand.randint(2):
        b = make_number(exp=exp,fionly=fionly,allowzero=allowbzero)
        if isinstance(b,uc.gummy):
            bx = b.x
        else:
            bx = b
        bu = 0
        bdof = float('inf')
    else:
        b,bx,bu,bdof,_,_ = make_gummy(unit=1,dof=dof,exp=exp,mpf=(not fionly),
                                      uexp=uexp,allowzero=allowbzero,
                                      allowlargeu=allowlargeu)
        
    ax = float(ax)
    bx = float(bx)
    au = float(au)
    bu = float(bu)
    x = f(ax,bx)
    d = df(ax,bx)
    da = float(d[0])
    db = float(d[1])
    u = np.sqrt((da*au)**2 + (db*bu)**2)
    if dof is None:
        dof = ((da*au)**4/adof + (db*bu)**4/bdof)
        if dof == 0:
            dof = float('inf')
        else:
            dof = u**4/dof
            
        if dof > uc.gummy.max_dof:
            dof = float('inf')
        if dof < 1:
            dof = 1
            
        dof = float(dof)
            
    g = f(a,b)
    
    if x == 0:
        assert abs(g.x) < 1e-15
    else:
        assert abs((g.x - x)/x) < 1e-14
        
    if u == 0:
        assert abs(u) < 1e-6
    else:
        assert abs((g.u - u)/u) < 1e-3
    
    if dof is not None and not np.isinf(dof):
        assert abs((g.dof - dof)/dof) < 0.01
    
    assert a.correlation(b) == 0
    
    if sim and x != 0 and abs(u/x) > 1e-10:
        uc.gummy.simulate([a,b,g])
        
        assert abs((g.xsim - x)/(g.u/np.sqrt(1e5))) < 5
        assert abs((g.u - g.usim)/g.u) < 0.1
        
        if g.correlation(a) > 0.1:
            assert abs((g.correlation_sim(a)-g.correlation(a))/g.correlation(a)) < 0.1
            
        if isinstance(b,uc.gummy) and g.correlation(b) > 0.1:
            assert abs((g.correlation_sim(b)-g.correlation(b))/g.correlation(b)) < 0.1
    
def test_add(n=1000,sim=False):
    def f(a,b):
        return a + b
    def df(ax,bx):
        return (1,1)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_sub(n=1000,sim=False):
    def f(a,b):
        return a - b
    def df(ax,bx):
        return (1,-1)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_mul(n=1000,sim=False):
    def f(a,b):
        return a*b
    def df(ax,bx):
        return (bx,ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_div(n=1000,sim=False):
    def f(a,b):
        return a/b
    def df(ax,bx):
        return (1/bx,-ax/bx**2)
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,allowbzero=False,allowlargeu=lu)
        
def test_pow(n=1000,sim=False):
    def f(a,b):
        return abs(a)**b
    def df(ax,bx):
        ax = abs(ax)
        return (bx*ax**(bx-1),np.log(ax)*ax**bx)
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,exp=0,allowazero=False,allowlargeu=lu)
        
def test_mod(n=1000,sim=False):
    def f(a,b):
        return a%b
    def df(ax,bx):
        return (1,np.sign(bx)*abs(ax//bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
    
def test_abs(n=1000,sim=False):
    def f(a,b):
        return abs(a)*abs(b)
    def df(ax,bx):
        return (abs(bx),abs(ax))
    if sim:
        lu = False
    else:
        lu = True
    for i in range(n):
        binary_func(f,df,sim=sim,allowlargeu=lu,allowazero=lu,
                    allowbzero=lu)
        
def test_neg(n=1000,sim=False):
    def f(a,b):
        return (-a)*b
    def df(ax,bx):
        return (-bx,-ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_pos(n=1000,sim=False):
    def f(a,b):
        return (+a)*b
    def df(ax,bx):
        return (bx,ax)
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_sincos(n=1000,sim=False):
    def f(a,b):
        return uc.sin(a) + uc.cos(b)
    def df(ax,bx):
        return (uc.cos(ax),-uc.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim)
        
def test_npsincos(n=1000,sim=False):
    def f(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
        
def test_ap1sincos(n=1000,sim=False):
    def f(a,b):
        return uc.gummy.apply(np.sin,np.cos,a) + uc.gummy.napply(np.cos,b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
        
def test_ap2sincos(n=1000,sim=False):
    def ff(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    def f(a,b):
        return uc.gummy.apply(ff,df,a,b)

    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True)
        
def test_apnsincos(n=1000,sim=False):
    def ff(a,b):
        return np.sin(a) + np.cos(b)
    def df(ax,bx):
        return (np.cos(ax),-np.sin(bx))
    def f(a,b):
        return uc.gummy.napply(ff,a,b)

    for i in range(n):
        binary_func(f,df,sim=sim,fionly=True,exp=0,allowlargeu=False)
        
def test_addxmul(n=1000):
    def f(*x):
        r = x[0]
        for i in range(len(x)-1):
            r += x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1)[0]
        a = f(*(k*[x]))
        b = k*x
        c = uc.gummy.apply(f,lambda *x: len(x)*[1],*(k*[x]))
        d = uc.gummy.napply(f,*(k*[x]))
        
        if b.x == 0:
            assert abs(a.x - b.x) < 1e-14
        else:
            assert abs((a.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert a.u == 0
            assert b.u == 0
        else:
            assert (a.u - b.u)/b.u < 1e-3
            assert a.correlation(b) == 1
            assert a.correlation(x) == 1
        
        if b.x == 0:
            assert abs(c.x - b.x) < 1e-14
        else:
            assert abs((c.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert c.u == 0
        else:
            assert (c.u - b.u)/b.u < 1e-3
            assert c.correlation(b) == 1
        
        if b.x == 0:
            assert abs(d.x - b.x) < 1e-14
        else:
            assert abs((d.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert d.u == 0
        else:
            assert abs((d.u - b.u)/b.u) < 1e-3
            assert d.correlation(b) == 1
        
def test_mulxpow(n=1000):
    def f(*x):
        r = x[0]
        for i in range(len(x)-1):
            r *= x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1)[0]
        a = f(*(k*[x]))
        b = x**k
        xx = x.x
        c = uc.gummy.apply(f,lambda *x: len(x)*[xx**(k-1)],*(k*[x]))
        d = uc.gummy.napply(f,*(k*[x]))
        
        if b.x == 0:
            assert abs(a.x - b.x) < 1e-14
        else:
            assert abs((a.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert a.u == 0
            assert b.u == 0
        else:
            assert abs((a.u - b.u)/b.u) < 1e-3
            assert a.correlation(b) == 1
        
        if k%2 == 0 and x.x < 0:
            assert a.correlation(x) == -1
        else:
            assert a.correlation(x) == 1
            
        if b.x == 0:
            assert abs(c.x - b.x) < 1e-14
        else:
            assert abs((c.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert c.u == 0
        else:
            assert abs((c.u - b.u)/b.u) < 1e-3
            assert c.correlation(b) == 1
        
        if b.x == 0:
            assert abs(d.x - b.x) < 1e-14
        else:
            assert abs((d.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert d.u == 0
        else:
            assert abs((d.u - b.u)/b.u) < 1e-3
            assert d.correlation(b) == 1

        
def test_mulxnpow(n=1000):
    def f(*x):
        r = 1/x[0]
        for i in range(len(x)-1):
            r *= 1/x[i+1]
        return r
        
    for j in range(n):
        k = rand.randint(10) + 2
        x = make_gummy(unit=1,allowzero=False,allowlargeu=False)[0]
        a = f(*(k*[x]))
        b = x**-k
        d = uc.gummy.napply(f,*(k*[x]))
        
        if b.x == 0:
            assert abs(a.x - b.x) < 1e-14
        else:
            assert abs((a.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert a.u == 0
            assert b.u == 0
        else:
            assert abs((a.u - b.u)/b.u) < 1e-3
            assert a.correlation(b) == 1
        
        if k%2 == 0 and x.x < 0:
            assert a.correlation(x) == 1
        else:
            assert a.correlation(x) == -1
        
        if b.x == 0:
            assert abs(d.x - b.x) < 1e-14
        else:
            assert abs((d.x - b.x)/b.x) < 1e-14
        if x.u == 0:
            assert d.u == 0
        else:
            assert abs((d.u - b.u)/b.u) < 1e-3
            assert d.correlation(b) == 1


# -*- coding: utf-8 -*-

# module budget

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
This module defines the Budget object which displays a table listing the 
uncertainty contributions to a gummy.
"""

import numpy as np
from .gummy import gummy
from .unit import one
from .exceptions import BudgetWarning
from .printing import PrettyPrinter


import warnings
from collections import OrderedDict

class Budget(PrettyPrinter):
    """
    A class that facilitates the creation of uncertainty budget tables.
    
    To display the table use the `Budget.html` or `Budget.latex`
    methods  in a console or notebook that supports this type of output
    or the python  built-in function to get a unicode table.

    The `Budget.tohtml` and `Budget.tolatex` methods can be used
    to get strings with the html or latex code.

    The `Budget.df` property can be used to retrieve a pandas `DataFrame`
    with the table.  Also `Budget.df_str`, `Budget.df_html` and
    `Budget.df_latex` return DataFrames with formatted strings as entries
    rather than numerical values.

    Parameters
    ----------
    y:  `gummy`
        the dependant variable

    xlist:   array_like of `gummy`
        The independent variables.  Warnings will be generated if the
        gummys in this list over determine `y` (that is if not all
        variables in this list can be treated as independent variables)
        or under determine `y` (that is if some variables
        contributing to the uncertainty in `y` are missing).

    uunit:  `str` or Unit, optional
        Unit to use to express the uncertainties.  This useful if you
        wish to express all uncertainties  as relative uncertainty unit
        (e.g. %).

    k, p: `float`, optional
        k or p values for the expanded uncertainty; do not specify both `k`
        and `p`; if neither are specified the the `k` and `p` values of `y`
        are used

    sort:  `bool`, optional
        Whether or not to sort the gummys in `xlist` by significance.

    columns: `list` of `str` or `None`, optional
        Allows the user to select the columns (and ordering of the columns)
        for display.  The available columns are:

        "component" or "name":  the names of the gummy, displayed by default

        "description":  description given in the description parameter list,
            displayed by default if the description parameter is not None

        "unit":  the unit of the gummy, displayed by default

        "value":  the x value of the gummy, displayed by default

        "u" or "uncertainty":  The uncertainty of the gummy.  This is the
            standard uncertainty except possible in the last row where an
            expanded uncertainty is displayed.  This column is displayed by
            default.

        "dof":  the degrees of freedom for the uncertainty, displayed by default
            if any uncertainty has finite degrees of freedom

        "type":  the uncertainty type, displayed by default if any gummy has
            a type defined

        "s" or "significance":  the sensitivity coefficient (below) multiplied
            by the standard uncertainty, displayed by default

        "d", "derivative" or "partial":  the partial derivative of the y gummy
            with resect to the gummy in that row

        "c" or "sensitivity coefficient":  the absolute value of "d"

        "custom":  value given in the custom parameter list, displayed by default
            if the custom parameter is not None

        The columns displayed can also be set with the `columns` property.

    column_names:  `dict` or `None`, optional
        Names to display as column headers, if this is `None` then the default
        names are used.  The dictionary should use as keys any of the column
        names listed above in the columns parameter description and as values
        the desired heading for this column.  The column names can also be set
        with the `column_names` property.

    show_subtotals: `bool`, optional
        If any uncertainty types are defined, the combined standard uncertainty
        for each type is displayed in the table.  This can also be changed by
        setting the `show_subtotals` attribute.  The default is `True`.

    show_expanded_u:  `bool` or `None`, optional
        Whether or not to display the expanded uncertainty in the last row.  If
        this is `None`, then the expanded uncertainty is displayed if ``y.k != 1``.
        This can also be changed by setting the `show_expanded_u` attribute.

    show_s:  `bool`, optional
        Whether or not to show the significance column.  This is ignored if the
        `columns` parameter is not `None`.  The default can be changed by setting
        the attribute class attribute `Budget.show_s`.  The default value is `True`.

    show_d:  `bool`, optional
        Whether or not to show the partial derivatives column.  This is ignored
        if columns is not None.  The default can be  changed by setting the
        attribute class attribute `Budget.show_d`.  The default is `True`.

    show_c:  `bool`, optional
        Whether or not to show the sensitivity coefficient column.  This is ignored
        if columns is not None.  The default can be changed by setting the
        attribute class attribute `Budget.show_c`.  Teh default value is `False`

    units_on_values:  `bool` or `None`
        If this is `True`, units are shown in the value and u columns and if `False`
         the units are in a separate column.  If `None` then the units are in a
         separate column unless  any gummy in `xlist` or `y` has a `uunit` defined.

    sim:  `bool`, optional
        If True, the combined uncertainty and partial derivatives will be calculated
        using Monte-Carlo data.  The default is `False`

    css:  `str` or `None`, optional
        A css header to be used when displaying the table in HTML format.  If this is
        `None` then `Budget.default_css` will be used.

    description:  `list` of `str` or `None`, optional
        An optional column of descriptions to be printed in the table.  This should
        be a description for `y` then for each `x`, and followed, optionally, by
        subtotal and expanded uncertainty descriptions.

    description_math_mode:  `bool`, optional
        If this is `False`, then when using a LaTeX format, the description is put
        in normal text mode rather than math mode.  The default is `False`

    custom:  `list` of `str` or `None`, optional
        An optional column of  additional information to be printed in the table.
        This should be a value for `y` then for each `x`, and followed, optionally,
        by subtotal and expanded uncertainty values.

    custom_heading:  `str` or `None`, optional
        A heading for the custom column.

    custom_math_mode:  `bool`
        If this is False, then when using a LaTeX format, the custom value is put
        in normal text mode rather than math mode.  The default value is `False`.

    solidus, mulsep, slashaxis:
        see gummy.solidus, gummy.mulsep and gummy.slashaxis
    """
    
    default_css = """
.budget_table {
    display: table;
}
    
.budget_row {
    display: table-row;
}

.budget_header_row {
    display: table-row;
    font-weight: bold;
}

.budget_first_col_header_cell
{
    display: table-cell;
    text-align: center;
    border-bottom: solid;
    border-width: thin;
    padding-left: 5px;
    padding-right: 5px;
}

.budget_header_cell {
    display: table-cell;
    text-align: center;
    border-bottom: solid;
    border-width: thin;
    padding-left: 5px;
    padding-right: 5px;
    
}

.budget_first_col_x_cell {
    display: table-cell;
    text-align: center;
    padding-left: 5px;
    padding-right: 5px;
    min-width: 8ex;
}

.budget_x_cell {
    display: table-cell;
    text-align: center;
    padding-left: 5px;
    padding-right: 5px;
    min-width: 8ex;
}
"""
    
    @staticmethod
    def _formatsig(x,fmt=None):
        if x < 10**-9:
            return '0'
        if x >= 0.01:
            return '{:.2f}'.format(x)
        if x >= 0.001:
            return '{:.3f}'.format(x)
        return Budget._format_float(x,nsig=1,fmt=fmt)
        
    @staticmethod
    def _format_float(x,fmt=None,nsig=None):
        g = gummy(x,abs(x))
        g.style = 'xf'
        if nsig is not None:
            g.nsig = nsig
        return g.tostring(fmt)
    
    units_on_values = None
    show_s = True
    show_d = False
    show_c = True
    default_columns = None
    default_column_names = None
        
    def __init__(self,y,xlist,uunit=None,units_on_values=None,
                 sort=True,solidus=None,mulsep=None,slashaxis=None,columns=None,
                 column_names=None,xnames=None,yname=None,show_subtotals=True,
                 show_expanded_u=None,description=None,description_math_mode=False,
                 custom=None,custom_heading=None,custom_math_mode=False,
                 show_s=None,show_d=None,show_c=None,css=None,k=None,p=None,
                 sim=False):
        from pandas import DataFrame
        self._DataFrame = DataFrame
        
        if solidus is None:
            self.solidus = gummy.solidus
        else:
            self.solidus = solidus
            
        if mulsep is None:
            self.mulsep = gummy.mulsep
        else:
            self.mulsep = mulsep
            
        if slashaxis is None:
            self.slashaxis = gummy.slashaxis
        else:
            self.slashaxis = slashaxis
            
        if show_s is not None:
            self.show_s = show_s
        if show_d is not None:
            self.show_d = show_d
        if show_c is not None:
            self.show_c = show_c
        
        self.sim = sim
        
        got_dof = np.isfinite(y.dof) or any(np.isfinite(g.dof) for g in xlist)
        got_unit = (y.unit is not one) or any(g.unit is not one for g in xlist)
        got_uunit = (((y.uunit is not None) or any(g.uunit is not None for g in xlist))
                     and uunit is None)
        got_tag = any(g.utype is not None for g in xlist)
            
        self._y = y.copy(formatting=True)
        if k is not None or p is not None:
            if k is not None and p is not None:
                raise ValueError('k and p may not both be specified')
            if k is not None:
                self._y.k = k
            else:
                self._y.p = p
            
        if uunit is not None:
            self._y.uunit = uunit
        self._yu = y.copy(formatting=True)
        self._yu.k = 1
        
        self.nx = len(xlist)
        
        ind = list(range(len(xlist)))
        
        if xnames is not None and len(xnames) != len(xlist):
            raise ValueError('len(xnames) != len(xlist)')
        if xnames is not None:
            xnames = [str(nm).strip() for nm in xnames]
            hxnames = ['<span><i>' + str(nm).strip() + '</i></span>' 
                       if len(nm) <= 1 else str(nm).strip() for nm in xnames]
            lxnames = [str(nm).strip() if len(nm) <= 1 else self.latex_norm(nm) for nm in xnames]
            
        else:
            xnames = []
            hxnames = []
            lxnames = []
            for i,x in enumerate(xlist):
                if x.name is None:
                    xnames.append('x[' + str(i+1) + ']')
                    hxnames.append('<span><i>x</i><sub>' + str(i+1) + '</sub></span>')
                    lxnames.append('x_{' + str(i+1) + '}')
                else:
                    xnames.append(str(x.name).strip())
                    if not isinstance(x.name,str) or len(x.name) > 1:
                        hxnames.append(str(x.name).strip())
                        lxnames.append(type(self).latex_norm(str(x.name).strip()))
                    else:
                        hxnames.append('<span><i>' + str(x.name).strip() + '</i></span>')
                        lxnames.append(str(x.name).strip())
                        
            
        if yname is None:
            if y.name is None:
                yname = 'y'
            else:
                yname = y.name
        self._yname = str(yname).strip()
        if len(yname) <= 1:
            self._hyname = '<span><i>' + str(yname).strip() + '</i></span>'
            self._lyname = str(yname).strip()
        else:
            self._hyname = str(yname).strip()
            self._lyname = type(self).latex_norm(str(yname).strip())
            
        
        if units_on_values is not None:
            self.units_on_values = units_on_values
        
        self.show_subtotals = show_subtotals
        self.show_expanded_u = show_expanded_u
        self.css = css
        
        self.uunit = uunit
        
        if description is None:
             self.ydescription = None
             self.yedescription = None
             self.tdescription = None
             description = (self.nx+1)*['']
        else:
            if len(description) < self.nx + 1:
                raise ValueError('description must have a length of at least len(xlist) + 1')
            self.ydescription = description[0]
            if len(description) > self.nx + 1:
                self.yedescription = description[-1]
                if len(description) > self.nx + 1:
                    self.tdescription = description[(self.nx + 1):-1]
                else:
                    self.tdescription = None
            else:
                self.yedescription = None
                self.tdescription = None
        self.description_math_mode = description_math_mode
        if custom is None:
            self.ycustom = None
            self.yecustom = None
            self.tcustom = None
            custom = (self.nx+1)*['']
        else:
            if len(custom) < self.nx + 1:
                raise ValueError('custom must have a length of at least len(xlist) + 1')
            self.ycustom = custom[0]
            if len(custom) > self.nx + 1:
                self.yecustom = custom[-1]
                if len(custom) > self.nx + 1:
                    self.tcustom = custom[(self.nx + 1):-1]
                else:
                    self.tcustom = None
            else:
                self.yecustom = None
                self.tcustom = None
        self.custom_math_mode = custom_math_mode
        if custom_heading is None:
            self.custom_heading = ''
        else:
            self.custom_heading = custom_heading
            
        x = xlist
            
        x = [g.copy(formatting=True) for g in x]
        for g in x:
            g.k = 1
            if uunit is not None:
                g.uunit = uunit
            
        if self.sim:
            cm = gummy.correlation_matrix_sim(x)
            yu = y.usim
        else:
            cm = gummy.correlation_matrix(x)
            yu = y._u
        svd = np.linalg.svd(cm,compute_uv=False)
        for v in svd:
            if v < 1e-6:
                warnings.warn('the x variables are over determined; they cannot all be taken as independant variables',BudgetWarning,stacklevel=2)
                break
            
        b = [y.correlation(z) for z in x]
        s = np.linalg.lstsq(cm,b,rcond=None)[0]
        tu = 0
        
        if self.sim:
            d = [p*yu/q.usim for p,q in zip(s,x)]  
            for i in range(len(x)):
                for j in range(len(x)):
                    tu += d[i]*d[j]*x[i].correlation(x[j])*x[i].usim*x[j].usim
        else:
            d = [p*yu/q._u for p,q in zip(s,x)]  
            for i in range(len(x)):
                for j in range(len(x)):
                    tu += d[i]*d[j]*x[i].correlation(x[j])*x[i]._u*x[j]._u
        tu = np.sqrt(tu)
        
        if self.sim:
            mlim = 0.05
        else:
            mlim = 0.001
        if np.abs((yu - tu)/yu) > mlim:
            warnings.warn('a source of uncertainty seems to be missing from the x variables',BudgetWarning,stacklevel=2)
        
        if sim:
            fsigs = [np.abs(i*j.usim/yu) for i,j in zip(d,x)]
        else:
            fsigs = [np.abs(i*j._u/yu) for i,j in zip(d,x)]
        fsc = list(np.abs(d))
        
        sigs = []
        html_sigs = []
        latex_sigs = []
        html_d = []
        latex_d = []
        str_d = []
        sc = []
        html_sc = []
        latex_sc = []
        fdof = []
        dof = []
        html_dof = []
        latex_dof = []
     
        for i,xi in enumerate(x):
            sigs.append(Budget._formatsig(fsigs[i]))
            html_sigs.append(Budget._formatsig(fsigs[i],fmt='html'))
            latex_sigs.append(Budget._formatsig(fsigs[i],fmt='latex'))
            str_d.append(Budget._format_float(d[i]))
            html_d.append(Budget._format_float(d[i],fmt='html'))
            latex_d.append(Budget._format_float(d[i],fmt='latex'))
            sc.append(Budget._format_float(fsc[i]))
            html_sc.append(Budget._format_float(fsc[i],fmt='html'))
            latex_sc.append(Budget._format_float(fsc[i],fmt='latex'))
            fdof.append(x[i]._dof)
            dof.append(gummy._dof_to_str(x[i].dof))
            html_dof.append(gummy._dof_to_str(x[i].dof,fmt='html'))
            latex_dof.append(gummy._dof_to_str(x[i].dof,fmt='latex'))
        
        a = np.array([x,ind,xnames,hxnames,lxnames,dof,html_dof,latex_dof,fdof,
                      sigs,html_sigs,latex_sigs,fsigs,str_d,html_d,latex_d,d,sc,
                      html_sc,latex_sc,fsc,description[1:(self.nx + 1)],custom[1:(self.nx + 1)]]).T

        self._dfx = DataFrame(a,columns=['x','i','xnames','hxnames','lxnames',
                                         'dof','html_dof','latex_dof','fdof',
                                          's','html_s','latex_s','fs','d',
                                          'html_d','latex_d','fd','sc','html_sc',
                                          'latex_sc','fsc','description','custom'])
        if sort:
            self._dfx.sort_values(by=['fs'],inplace=True,ascending=False)
            
        stags = set([])
        for r in xlist:
            if r.utype is not None:
                stags.add(r.utype)
        self.type = OrderedDict()
        if stags:
            lst = list(stags)
            lst.sort()
            for t in lst:
                u = self._y.ufrom(t,sim=self.sim)
                g = gummy(self._y.x,u=u,unit=self._y.unit)
                if self._y.uunit is not None:
                    g.uunit = self._y.uunit
                self.type[t] = (u,self._y.doffrom(t),g)               

        if self.units_on_values is None:
           self.units_on_values = got_uunit
            
        if columns is None:
            if self.default_columns is not None:
                self.columns = self.default_columns
            else:
                cols = ['name']
                if self.ydescription is not None:
                    cols.append('description')
                if got_unit and not self.units_on_values:
                    cols.append('unit')
                cols.append('value')
                cols.append('u')
                if got_dof:
                    cols.append('dof')
                if got_tag:
                    cols.append('type')
                if self.show_d:
                    cols.append('d')
                if self.show_c:
                    cols.append('c')
                if self.show_s:
                    cols.append('s')
                if self.ycustom is not None:
                    cols.append('custom')
            
            self.columns = cols
        else:
            self.columns = columns
            
        if column_names is None and self.default_column_names is not None:
            self.column_names = self.default_column_names
        else:
            self.column_names = column_names
        
    @property
    def columns(self):
        """
        `list` of `str` or `None`

        Allows the user to select the columns (and ordering of the columns)
        for display.  The available columns are:

        "component" or "name":  the names of the gummy, displayed by default

        "description":  description given in the description parameter list,
            displayed by default if the description parameter is not None

        "unit":  the unit of the gummy, displayed by default

        "value":  the x value of the gummy, displayed by default

        "u" or "uncertainty":  The uncertainty of the gummy.  This is the
            standard uncertainty except possible in the last row where an
            expanded uncertainty is displayed.  This column is displayed by
            default.

        "dof":  the degrees of freedom for the uncertainty, displayed by default
            if any uncertainty has finite degrees of freedom

        "type":  the uncertainty type, displayed by default if any gummy has
            a type defined

        "s" or "significance":  the sensitivity coefficient (below) multiplied
            by the standard uncertainty, displayed by default

        "d", "derivative" or "partial":  the partial derivative of the y gummy
            with resect to the gummy in that row

        "c" or "sensitivity coefficient":  the absolute value of "d"

        "custom":  value given in the custom parameter list, displayed by default
            if the custom parameter is not None
        """
        return self._columns
    @columns.setter
    def columns(self,cols):
        self._columns = []
        for c in cols:
            c = c.strip().lower()
            if c == 'name' or c == 'names' or c == 'component' or c == 'components':
                self._columns.append('component')
            elif c == 'unit' or c == 'units':
                self._columns.append('unit')
            elif c == 'value' or c == 'values':
                self._columns.append('value')
            elif c == 'u' or c == 'uncertainty' or c == 'uncertainties':
                self._columns.append('u')
            elif c == 'dof':
                self._columns.append('dof')
            elif c == 'type' or c == 'tag':
                self._columns.append('type')
            elif c == 's' or c == 'significance' or c == 'signifincances':
                self._columns.append('s')
            elif c == 'd' or c == 'derivative' or c == 'derivatives' or c == 'partial' or c == 'partials':
                self._columns.append('d')
            elif c == 'c' or c == 'sensitivity coefficient' or c == 'sensitivity coefficients':
                self._columns.append('c')
            elif c == 'description':
                self._columns.append('description')
            elif c == 'custom':
                self._columns.append('custom')
            else:
                raise ValueError('column name ' + str(c) + ' is unrecognized')
                
    @property
    def column_names(self):
        """
       `dict`

        Names to display as column headers.  The dictionary should use as keys
        any of the column names listed above in the columns parameter
        description and as values the desired heading for this column.
        """
        return self._column_names
    @column_names.setter
    def column_names(self,names):
        self._column_names = {}
        if names is None:
            return
        for c,v in names.items():
            c = c.strip().lower()
            v = v.strip()
            if c == 'name' or c == 'names' or c == 'component' or c == 'components':
                self._column_names['component'] = v
            elif c == 'unit' or c == 'units':
                self._column_names['unit'] = v
            elif c == 'value' or c == 'values':
                self._column_names['value'] = v
            elif c == 'u' or c == 'uncertainty' or c == 'uncertainties':
                self._column_names['u'] = v
            elif c == 'dof':
                self._column_names['dof'] = v
            elif c == 'type' or c == 'tag':
                self._column_names['type'] = v
            elif c == 's' or c == 'significance' or c == 'significances':
                self._column_names['s'] = v
            elif c == 'd' or c == 'derivative' or c == 'derivatives' or c == 'partial' or c == 'partials':
                self._column_names['d'] = v
            elif c == 'c' or c == 'sensitivity coefficient' or c == 'sensitivity coefficients':
                self._column_names['c'] = v
            elif c == 'description':
                self._columns.names['description'] = v
            elif c == 'custom':
                self._columns.names['custom'] = v
            else:
                raise ValueError('original column name ' + str(c) + ' is unrecognized')
                
    @property
    def k(self):
        """
        Gets or sets the `k` value for the expanded uncertainty of `y`.
        """
        return self._y.k
    @k.setter
    def k(self,value):
        self._y.k = value
        
    @property
    def p(self):
        """
        Gets or sets the `p` value for the expanded uncertainty of `y`.
        """
        return self._y.p
    @p.setter
    def p(self,value):
        self._y.p = value
            
    @property
    def df(self):
        """
        read-only

        Returns a Panda's DataFrame with the the budget table.
        """
        self._update()
        return self._df
        
    @property
    def df_str(self):
        """
        read-only

        Returns a Panda's DataFrame with the the budget table, with
        all values displayed as strings.
        """
        self._update()
        return self._df_str
        
    @property
    def df_html(self):
        """
        read-only

        Returns a Panda's DataFrame with the the budget table, with
        entries displayed using HTML.
        """
        self._update()
        return self._df_html
        
    @property
    def df_latex(self):
        """
        read-only

        Returns a Panda's DataFrame with the the budget table, with all
        entries displayed using LaTeX.
        """
        self._update()
        return self._df_latex
        
    def _update(self):        
        show_exu = self.show_expanded_u or (self.show_expanded_u is None and self._y.k != 1)
       
        self._df = self._DataFrame()
        self._df_str = self._DataFrame()
        self._df_html = self._DataFrame()
        self._df_latex = self._DataFrame()
                        
        for c in self._columns:
            if c == 'component':
                nms = list(self._dfx['xnames'])
                hnms = list(self._dfx['hxnames'])
                lnms = list(self._dfx['lxnames'] )
                if self.show_subtotals:
                    for t in self.type.keys():
                        nms.append('Combined type ' + str(t))
                        hnms.append('<span><i>u<sub>c</sub></i> type ' + str(t))
                        lnms.append('u_{c} \\text{ type ' + str(t) + '}')
                nms.append(self._yname)
                hnms.append(self._hyname)
                lnms.append(self._lyname)
                if show_exu:
                    if self.sim and not self._y._set_k:
                        k = self._y.ksim
                    else:
                        k = self._y.k
                    nms.append('Uc at k = ' + gummy._k_to_str(k))
                    hnms.append('<span><i>U<sub>c<sub></i> at <i>k</i> = ' + gummy._k_to_str(k) + '</span>')
                    lnms.append('U_c ' + type(self).latex_norm(' at ') + ' k = ' + gummy._k_to_str(k))
                    
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = 'Component'
                    cnml = type(self).latex_norm('Component')
                else:
                    cnml = cnm
                    
                self._df[cnm] = nms
                self._df_str[cnm] = nms
                self._df_html[cnm] = hnms
                self._df_latex[cnml] = lnms
                
            if c == 'unit':
                sym = []
                hsym = []
                lsym = []
                for r in self._dfx['x']:
                    if isinstance(r,gummy):
                        sym.append(r.unit.tostring(solidus=self.solidus,mulsep=self.mulsep))
                        hsym.append(r.unit.tostring(fmt='html',solidus=self.solidus,mulsep=self.mulsep))
                        lsym.append(r.unit.tostring(fmt='latex',solidus=self.solidus,mulsep=self.mulsep))
                    else:
                        sym.append('')
                        hsym.append('')
                        lsym.append('')
                        
                if isinstance(self._y,gummy):
                    yun = self._y.unit.tostring(solidus=self.solidus,mulsep=self.mulsep)
                    hyun = self._y.unit.tostring(fmt='html',solidus=self.solidus,mulsep=self.mulsep)
                    lyun = self._y.unit.tostring(fmt='latex',solidus=self.solidus,mulsep=self.mulsep)
                else:
                    yun = ''
                    hyun = ''
                    lyun = ''
                
                n = 1
                if self.show_subtotals:
                    n += len(self.type.values())
                            
                if show_exu:
                    n += 1

                sym += n*[yun]
                hsym += n*[hyun]
                lsym += n*[lyun]
                    
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = 'Unit'
                    cnml = type(self).latex_norm('Unit')
                self._df[cnm] = sym
                self._df_str[cnm] = sym
                self._df_html[cnm] = hsym
                self._df_latex[cnml] = lsym
                
            if c == 'value':
                if self.units_on_values:
                    if self.sim:
                        xst = 'xsim'
                    else:
                        xst = 'x'
                else:
                    if self.sim:
                        xst = 'xfsim'
                    else:
                        xst = 'xf'
                    
                fvalues = [g.x for g in self._dfx['x']]
                values = [g.tostring(style=xst,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                hvalues = [g.tostring(fmt='html',style=xst,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                lvalues = [g.tostring(fmt='latex',style=xst,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                
                if self.show_subtotals:
                    for t in self.type.values():
                        fvalues.append(None)
                        values.append('')
                        hvalues.append('')
                        lvalues.append('')
                        
                if self.sim:
                    fvalues.append(self._y.xsim)
                else:
                    fvalues.append(self._y.x)
                values.append(self._y.tostring(style=xst))
                hvalues.append(self._y.tostring(fmt='html',style=xst))
                lvalues.append(self._y.tostring(fmt='latex',style=xst))
                
                if show_exu:
                    fvalues.append(None)
                    values.append('')
                    hvalues.append('')
                    lvalues.append('')
                    
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = 'Value'
                    cnml = type(self).latex_norm('Value')
                self._df[cnm] = fvalues
                self._df_str[cnm] = values
                self._df_html[cnm] = hvalues
                self._df_latex[cnml] = lvalues
                
            if c == 'u':
                if self.units_on_values:
                    ust_g = 'u'
                    if self.sim:
                        ust = 'usim'
                    else:
                        ust = 'u'
                else:
                    ust_g = 'uf'
                    if self.sim:
                        ust = 'ufsim'
                    else:
                        ust = 'uf'
                    
                fu = [g.u for g in self._dfx['x']]
                u = [g.tostring(style=ust,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                hu = [g.tostring(fmt='html',style=ust,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                lu = [g.tostring(fmt='latex',style=ust,solidus=self.solidus,mulsep=self.mulsep) for g in self._dfx['x']]
                
                if self.show_subtotals:
                    for t in self.type.values():
                        fu.append(t[2].U)
                        u.append(t[2].tostring(style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                        hu.append(t[2].tostring(fmt='html',style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                        lu.append(t[2].tostring(fmt='latex',style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                            
                fu.append(self._yu.U)
                u.append(self._yu.tostring(style=ust,solidus=self.solidus,mulsep=self.mulsep))
                hu.append(self._yu.tostring(fmt='html',style=ust,solidus=self.solidus,mulsep=self.mulsep))
                lu.append(self._yu.tostring(fmt='latex',style=ust,solidus=self.solidus,mulsep=self.mulsep))
                if show_exu:
                    if self.sim:
                        g = gummy(self._y.x,self._y.ksim*self._y.usim,unit=self._y.unit)
                    else:
                        g = self._y
                    fu.append(g.U)
                    u.append(g.tostring(style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                    hu.append(g.tostring(fmt='html',style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                    lu.append(g.tostring(fmt='latex',style=ust_g,solidus=self.solidus,mulsep=self.mulsep))
                    
                cnm = cnmh = self.column_names.get(c)
                if cnm is None:
                    if self.uunit is None or self.units_on_values:
                        cnm = cnml = 'u'
                        cnmh = '<span><i>u</i></span>'
                    else:
                        if self.slashaxis:
                            cnm = 'u\u2009/\u2009' + self._y.uunit.tostring(solidus=self.solidus,mulsep=self.mulsep).strip()
                            cnml = 'u\\,/\\,' + self._y.uunit.tostring(fmt='latex',solidus=self.solidus,mulsep=self.mulsep).strip()
                            cnmh = '<span><i>u</i>&thinsp;/&thinsp;' + self._y.uunit.tostring(fmt='html',solidus=self.solidus,mulsep=self.mulsep).strip() + '</span>'
                        else:
                            cnm = 'u\u2009(' + self._y.uunit.tostring(solidus=self.solidus,mulsep=self.mulsep).strip() + ')'
                            cnml = 'u\\,(' + self._y.uunit.tostring(fmt='latex',solidus=self.solidus,mulsep=self.mulsep).strip() + ')'
                            cnmh = '<span><i>u</i>&thinsp;(' + self._y.uunit.tostring(fmt='html',solidus=self.solidus,mulsep=self.mulsep).strip() + ')</span>'
                self._df[cnm] = fu
                self._df_str[cnm] = u
                self._df_html[cnmh] = hu
                self._df_latex[cnml] = lu
                
            if c == 'dof':                
                fdof = list(self._dfx['fdof'])
                dof = list(self._dfx['dof'])
                hdof = list(self._dfx['html_dof'])
                ldof = list(self._dfx['latex_dof'] )
                if self.show_subtotals:
                    for t in self.type.values():
                        fdof.append(t[1])
                        dof.append(gummy._dof_to_str(t[1]))
                        hdof.append(gummy._dof_to_str(t[1],'html'))
                        ldof.append(gummy._dof_to_str(t[1],'latex'))
                fdof.append(self._y.dof)
                dof.append(gummy._dof_to_str(self._y.dof))
                hdof.append(gummy._dof_to_str(self._y.dof,fmt='html'))
                ldof.append(gummy._dof_to_str(self._y.dof,fmt='latex'))
                if show_exu:
                    fdof.append(None)
                    dof.append('')
                    hdof.append('')
                    ldof.append('')
                    
                cnm = self.column_names.get(c)
                if cnm is None:
                    self._df['DoF'] = fdof
                    self._df_str['DoF'] = dof
                    self._df_html['<span><i>&nu;<sub>eff</sub></i></span>'] = hdof
                    self._df_latex[r'\nu_{eff}'] = ldof
                else:
                    self._df[cnm] = fdof
                    self._df_str[cnm] = dof
                    self._df_html[cnm] = hdof
                    self._df_latex[cnm] = ldof
                
            if c == 'type':
                tags = []
                for r in self._dfx['x']:
                    if r.utype is not None:
                        tags.append(str(r.utype))
                    else:
                        tags.append('')
                        
                for t in self.type.keys():
                    if self.show_subtotals:
                        tags.append(str(t))                
                
                tags.append('')
                if show_exu:
                    tags.append('')
            
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = 'Type'
                    cnml = type(self).latex_norm('Type')
                self._df[cnm] = tags
                self._df_str[cnm] = tags
                self._df_html[cnm] = tags
                self._df_latex[cnml] = [type(self).latex_norm(t) for t in tags]
                
            if c == 's':                
                fs = list(self._dfx['fs'])
                s = list(self._dfx['s'])
                hs = list(self._dfx['html_s'])
                ls = list(self._dfx['latex_s'])
                if self.show_subtotals:
                    for v in self.type.values():
                        if self.sim:
                            yu = self._y.usim
                        else:
                            yu = self._y._u
                        fs.append(v[0]/yu)
                        s.append(Budget._formatsig(v[0]/yu))
                        hs.append(Budget._formatsig(v[0]/yu,'html'))
                        ls.append(Budget._formatsig(v[0]/yu,'latex'))
                fs.append(None)
                s.append('')
                hs.append('')
                ls.append('')
                if show_exu:
                    fs.append(None)
                    s.append('')
                    hs.append('')
                    ls.append('')
                    
                cnm = self.column_names.get(c)
                if cnm is None:
                    self._df['s'] = fs
                    self._df_str['s'] = s
                    self._df_html['<span><i>s</i></span>'] = hs
                    self._df_latex['s'] = ls
                else:
                    self._df[cnm] = fs
                    self._df_str[cnm] = s
                    self._df_html[cnm] = hs
                    self._df_latex[cnm] = ls
                
            if c == 'd':                
                fd = list(self._dfx['fd'])
                d = list(self._dfx['d'])
                hd = list(self._dfx['html_d'])
                ld = list(self._dfx['latex_d'] )
                if self.show_subtotals:
                    for v in self.type.values():
                        fd.append(None)
                        d.append('')
                        hd.append('')
                        ld.append('')
                fd.append(None)
                d.append('')
                hd.append('')
                ld.append('')
                if show_exu:
                    fd.append(None)
                    d.append('')
                    hd.append('')
                    ld.append('')
                    
                cnm = self.column_names.get(c)
                if cnm is None:
                    self._df['dy/dx'] = fd
                    self._df_str['dy/dx'] = d
                    self._df_html['<span>&part;<i>y</i>/&part;<i>x</i></span>'] = hd
                    self._df_latex[r'\frac{\partial y}{\partial x}'] = ld
                else:
                    self._df[cnm] = fd
                    self._df_str[cnm] = d
                    self._df_html[cnm] = hd
                    self._df_latex[cnm] = ld
                
            if c == 'c':                
                fc = list(self._dfx['fsc'])
                cc = list(self._dfx['sc'])
                hc = list(self._dfx['html_sc'])
                lc = list(self._dfx['latex_sc'])
                if self.show_subtotals:
                    for v in self.type.values():
                        fc.append(None)
                        cc.append('')
                        hc.append('')
                        lc.append('')
                fc.append(None)
                cc.append('')
                hc.append('')
                lc.append('')
                if show_exu:
                    fc.append(None)
                    cc.append('')
                    hc.append('')
                    lc.append('')
                    
                cnm = self.column_names.get(c)
                if cnm is None:
                    self._df['|dy/dx|'] = fc
                    self._df_str['|dy/dx|'] = cc
                    self._df_html['|<span>&part;<i>y</i>/&part;<i>x</i></span>|'] = hc
                    self._df_latex[r'\left\lvert\frac{\partial y}{\partial x} \right\rvert'] = lc
                else:
                    self._df[cnm] = fc
                    self._df_str[cnm] = cc
                    self._df_html[cnm] = hc
                    self._df_latex[cnm] = lc
                    
            if c == 'description':
                d = list(self._dfx['description'])
                if self.show_subtotals:
                    if (self.tdescription is not None and 
                        len(self.tdescription) == len(self.type.values())):
                        d += self.tdescription
                    else:
                        d += len(self.type.values())*['']
                d += [self.ydescription]
                if show_exu:
                    if self.yedescription is not None:
                        d += [self.yedescription]
                    else:
                        d += ['']
                
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = 'Description'
                    cnml = type(self).latex_norm('Description')
                self._df[cnm] = d
                self._df_str[cnm] = d
                self._df_html[cnm] = d
                if not self.description_math_mode:
                    self._df_latex[cnml] = [type(self).latex_norm(i) if i != '' 
                                            else '' for i in d]
                else:
                    self._df_latex[cnml] = d
                    
            if c == 'custom':
                d = list(self._dfx['custom'])
                if self.show_subtotals:
                    if (self.tcustom is not None and 
                        len(self.tcustom) == len(self.type.values())):
                        d += self.tcustom
                    else:
                        d += len(self.type.values())*['']
                d += [self.ycustom]
                if show_exu:
                    if self.yecustom is not None:
                        d += [self.yecustom]
                    else:
                        d += ['']
                
                cnm = cnml = self.column_names.get(c)
                if cnm is None:
                    cnm = self.custom_heading
                    if self.custom_heading == '':
                        cnml = ''
                    else:
                        cnml = type(self).latex_norm(self.custom_heading)
                self._df[cnm] = d
                self._df_str[cnm] = d
                self._df_html[cnm] = d
                if not self.custom_math_mode:
                    self._df_latex[cnml] = [type(self).latex_norm(i) if i != '' 
                                            else '' for i in d]
                else:
                    self._df_latex[cnml] = d
                        
        
    def tostring(self,fmt='unicode'):
        """
        Returns a string representation of the budget table

        Parameters
        ----------
        fmt: {'unicode','html','latex','ascii'}, optional
            encoding for the output.  The default is 'unicode'.
        """
        fmt = fmt.strip().lower()
        
        if fmt == 'unicode':
            return self.df_str.to_string(index=False)
        if fmt == 'html':
            return self._tohtml()
        if fmt == 'latex':
            return self._tolatex()
        if fmt == 'ascii':
            raise ValueError('fmt ascii is not available for Budget')
        raise ValueError('fmt ' + str(fmt) + ' is not valid')
        
    def _tolatex(self):
        self._update()
        nc = len(self._df_latex.columns)
        if self.show_subtotals:
            nt = len(self.type)
        else:
            nt = 0
        if self.nx == 0 or nc == 0:
            return ''
        ndf = len(self._df_latex.index)
        
        txt = r'\begin{array}{ '
        for i in range(nc):
            txt += 'c '
        txt += '}\n'
        
        aa = False
        for c in self._df_latex.columns:
            if aa:
                txt += ' & '
            else:
                aa = True
            txt += c
        txt += r' \\'
        txt += '\n'
        txt += r'\hline'
        txt += '\n'
       
        for i in range(ndf):
            aa = False
            for j in range(0,nc):
                v = self._df_latex.iloc[i,j]
                v = v.strip()
                if aa:
                    txt += ' & '
                else:
                    aa = True
                if v != '':
                    txt += v
            txt += r' \\'
            txt += '\n'
            if i == self.nx - 1 or i == self.nx+nt-1 or i == ndf-2:
                txt += r'\hline'
                txt += '\n'
        txt += '\end{array}'
        return txt

    def _tohtml(self):
        self._update()
        nx = self.nx
        nc = len(self._df_html.columns)
        if self.show_subtotals:
            nt = len(self.type)
        else:
            nt = 0
        if nx == 0 or nc == 0:
            return ''
        ndf = len(self._df_html.index)
            
        txt = '<div>\n<style>'
        if self.css is not None:
            txt += self.css
        else:
            txt += self.default_css
        txt += '</style>\n<div class="budget_table">\n    <div class="budget_header_row">\n'
        for c in self._df_html.columns:
            txt += '        <div class = "budget_header_cell">' + c + '</div>\n'
        txt += '    </div>\n'
      
        for i in range(ndf):
            txt += '    <div class="budget_row">\n'
            for j in range(0,nc):
                txt += '        <div class="'
                if i < nx - 1:
                    if j == 0:
                        txt += 'budget_first_col_x_cell'
                    else:
                        txt += 'budget_x_cell'
                elif i == nx - 1:
                    if j == 0:
                        txt += 'budget_first_col_header_cell'
                    else:
                        txt += 'budget_header_cell'
                elif i < (nx+nt-1):
                    if j == 0:
                        txt += 'budget_first_col_x_cell'
                    else:
                        txt +=  'budget_x_cell'
                elif i == (nx+nt-1):
                    if j == 0:
                        txt += 'budget_first_col_header_cell'
                    else:
                        txt += 'budget_header_cell'
                elif i == ndf-1:
                    if j == 0:
                        txt += 'budget_first_col_x_cell'
                    else:
                        txt += 'budget_x_cell'
                else:
                    if j == 0:
                        txt += 'budget_first_col_header_cell'
                    else:
                        txt += 'budget_header_cell'
                v = self._df_html.iloc[i,j]
                if v == '':
                    txt += '">  </div>\n'
                else:
                    txt += '">' + self._df_html.iloc[i,j] + '</div>\n'
            txt += '    </div>\n'
            
        txt += '</div>\n</div>'
        return txt



# -*- coding: utf-8 -*-

# module nummy

# Copyright (C) 2019 National Research Council Canada
# Author:  Harold Parks

# This file is part of MetroloPy.

# MetroloPy is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.

# MetroloPy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more 
# details.

# You should have received a copy of the GNU General Public License along with 
# MetroloPy. If not, see <http://www.gnu.org/licenses/>.

"""
The nummy object defined here was created as a super-class for gummy,
integrating the Monte-Carlo uncertainty propagation code in the distributions
module with the ummy object.  
"""
import numpy as np
from .ummy import ummy
from .distributions import (Distribution,TDist,NormalDist,MultivariateElement,
                            MultivariateDistribution,MultiNormalDist,MultiTDist)
from math import isinf,isfinite,isnan,sqrt

def _bop(f,npf,s,b):
    if not nummy._mcpropagate:
        return f
        
    if isinstance(b,nummy):
        f._dist = Distribution.apply(npf,s._dist,b._dist)
    else:
        f._dist = Distribution.apply(npf,s._dist,b)
    return f

def _rbop(f,npf,s,b):
    if not nummy._mcpropagate:
        return f
    
    f._dist = Distribution.apply(npf,b,s._dist)
    return f

def _uop(f,npf,s):
    if not nummy._mcpropagate:
        return f
        
    f._dist = Distribution.apply(npf,s._dist)
    return f


class MetaNummy(type):
    # Use a metaclass to define some "classproperties" for nummy   
    @property
    def mcpropagate(cls):
        """
        Setting this property to `False` turns of the code for Monte-Carlo
        uncertainty propagation.  This property should only be set once, before
        any gummy instances are created.  Turning `mcpropagate` off then on again
        may have unpredictable results.
        """
        return nummy._mcpropagate
    @mcpropagate.setter
    def mcpropagate(cls,value):
        nummy._mcpropagate = bool(value)
        
    @property
    def cimethod(cls):
        """
        str in {'shortest', 'symmetric'}
        
        Get or set the method for calculating the confidence interval from 
        Monte-Carlo data.  If this property is set at the class level, it will
        change the default `cimethod` value for new gummys but will not affect
        gummys that have already been created.
        
        Can be set either to the string 'shortest' or the string 'symmetric'.
        This property gets or sets the method for calculating confidence
        intervals from Monte-Carlo data.  
        
        If it is set to 'shortest', the confidence interval will be taken to be 
        the shortest interval that includes the desired fraction of the probability 
        distribution.  
        
        If it is set to 'symmetric', then the confidence interval will be set so 
        that, for n Monte-Carlo samples and a coverage probability of `p`, then
        `n`*(1-`p`)/2 samples lie below the lower limit of the confidence interval
        and the same number of samples lie above the upper limit of the confidence 
        interval.
        """
        return nummy._cimethod
    @cimethod.setter
    def cimethod(cls,value):
        value = value.lower().strip()
        if value not in ['shortest','symmetric']:
            raise ValueError('cimethod ' + str(value) + ' is not recognized')
        nummy._cimethod = value
        
    @property
    def bayesian(cls):
        """
        `bool`
        
        Read/write at the class level, but read-only at the instance level.
        The default value is `False`; this should only be changed once at the
        beginning of the session.  This property affects how the level of 
        confidence `p` (sometimes called coverage probability) of an expanded
        uncertainty is related to the coverage factor `k` for a gummy based on
        data with finite degrees of freedom.

        Standard uncertainties are often based on the standard deviation of a set
        of measurements (and the assumption that these measurements are drawn
        from a normally distributed population).  Traditionally (e.g. the GUM
        2008 edition) the standard uncertainty is taken to be the standard
        deviation of the mean (s/sqrt(n), where s is the sample standard deviation
        and n is the number of measurements).  However there is some "extra
        uncertainty" because the sample standard devation not exactly equal to
        the population standard deviation.  This is taken into account by using
        a Student's t distribution to calculate the expanded uncertainty.  However
        it has been pointed out, by those who advocate a Bayesian point of view,
        that the probability distribution for the measurand here is best described
        by a shifted and scaled Student's t distribution.  So the standard
        uncertainty should be the standard deviation of this distribution which
        is s*sqrt{(n-1)/[n*(n-3)]}.  Thus

        u(bayesian) = [dof/(dof - 2)]*u(traditional)

        where dof = n - 1 and the "extra uncertainty" is incorporated directly
        into the standard uncertainty.
        
        Example
        -------
        >>> gummy.bayesian = True
        >>> g = gummy(1,0.03,dof=5)
        >>> g.bayesian
        True
        """
        return nummy._bayesian
    @bayesian.setter
    def bayesian(cls,v):
        nummy._bayesian = bool(v)
    
class nummy(ummy,metaclass=MetaNummy):
    #  This class is not intended to be used directly and was created to contain
    #  the code that the gummy object uses for Monte-Carlo uncertainty propagation.

    _mcpropagate = True
    _cimethod = 'shortest'
    _bayesian = False # see the MetaNummy bayesian property
    
    def __init__(self,x,u=0,dof=float('inf'),utype=None,name=None):
        self._bayesian = nummy._bayesian
        
        if isinstance(x,ummy):
            self._copy(x,self,formatting=False)
            return
        
        self.name = name
        if isinstance(x,Distribution):
            if x._used or not x.isindependent:
                raise ValueError('Distribution instances may only be used as the x parameter of a new gummy only if they\nrepresent independant variables and have not previously been used with another gummy')                
            x._used = True
            if isinstance(x,MultivariateElement):
                raise TypeError('a MultivariateElement may not be used in a gummy initializer\nuse the create static method with a MultivariateDistribution')
            
            self._dist = x
            
            if hasattr(x,'dof'):
                if nummy._bayesian:
                    u = float(x.u())*x.dof/(x.dof-2)
                    dof = float('inf')
                else:
                    u = float(x.u())
                    dof = x.dof
            else:
                dof = float('inf')
                u = float(x.u())
                
            super().__init__(x.x(),u=u,dof=dof,utype=utype)
            
        else:
            super().__init__(x,u=u,dof=dof,utype=utype)
            x = float(x)
            u = float(self._u)
            if self._u == 0 or isnan(x) or isinf(x) or isnan(u):
                self._dist = x
                return
            dof = self._dof
            if isinf(dof):
                self._dist = NormalDist(x,u)
            else:
                if nummy._bayesian:
                    self._dist = TDist(x,u*(dof-2)/dof,dof)
                    self._dof = float('inf')
                else:
                    self._dist = TDist(x,u,dof)
            
    @property
    def distribution(self):
        """
        read-only

        Returns ths `Distribution` instance associated with the gummy.
        """
        return self._dist
    
    @property
    def dof(self):
        """
        float, read-only

        Returns the number or degrees of freedom that the uncertainty of the 
        gummy is based on.  If `gummy.bayesian` is set to `False`, then the Welch-
        Satterthwaite approximation is used to calculate the effective number
        of degrees of freedom for gummys that result from an operation between
        two or more other gummys.  A version of the Welch-Satterthwaite 
        approximation that takes into account correlations is used here, see
        [R. Willink, Metrologia, 44, 340 (2007)].  If `gummy.bayesian` is `True`
        then gummys that are the result from an opertaion between other gummys
        will always have dof = float('inf').
        """
        if isinf(self._dof) and hasattr(self._dist,'dof'):
            dof = self._dist.dof
        else:
            dof = self._dof
        if dof < 1:
            # Occasionally correlations can result in a dof less than 1;
            # see the ummy._get_dof method.
            return 1
        return dof
    
    @property
    def bayesian(self):
        """
        `bool`

        Read/write at the class level, but read-only at the instance level.
        The default value is `False`; this should only be changed once at the
        beginning of the session.  This property affects how the level of
        confidence `p` (sometimes called coverage probability) of an expanded
        uncertainty is related to the coverage factor `k` for a gummy based on
        data with finite degrees of freedom.

        Standard uncertainties are often based on the standard deviation of a set
        of measurements (and the assumption that these measurements are drawn
        from a normally distributed population).  Traditionally (e.g. the GUM
        2008 edition) the standard uncertainty is taken to be the standard
        deviation of the mean (s/sqrt(n), where s is the sample standard deviation
        and n is the number of measurements).  However there is some "extra
        uncertainty" because the sample standard devation not exactly equal to
        the population standard deviation.  This is taken into account by using
        a Student's t distribution to calculate the expanded uncertainty.  However
        it has been pointed out, by those who advocate a Bayesian point of view,
        that the probability distribution for the measurand here is best described
        by a shifted and scaled Student's t distribution.  So the standard
        uncertainty should be the standard deviation of this distribution which
        is s*sqrt{(n-1)/[n*(n-3)]}.  Thus

        u(bayesian) = [dof/(dof - 2)]*u(traditional)

        where dof = n - 1 and the "extra uncertainty" is incorporated directly
        into the standard uncertainty.

        Example
        -------
        >>> gummy.bayesian = True
        >>> g = gummy(1,0.03,dof=5)
        >>> g.bayesian
        True
        """
        return self._bayesian
        
    @staticmethod
    def simulate(nummys,n=100000,ufrom=None):
        if ufrom is not None:
            ufrom = [g._dist for g in ufrom]
        if isinstance(nummys,nummy):
            nummys = [nummys]
        Distribution.simulate([g.distribution for g in nummys],n,ufrom)
        
    @property
    def simdata(self):
        """
        `numpy.ndarray`, read-only

        Returns an array containing the Monte-Carlo simulation data.  A 
        `NoSimulatedDataError` is raised if no Monte-Carlo data is available.
        """
        if self._u == 0:
            raise TypeError('simulated data is not available from a constant')
        return self.distribution.simdata
        
    @property
    def simsorted(self):
        """
        `numpy.ndarray`, read-only

        Returns a sorted array containing the Monte-Carlo simulation data.  A 
        `NoSimulatedDataError` is raised if no Monte-Carlo data is available.
        """
        if self._u == 0:
            raise TypeError('simulated data is not available from a constant')
        return self.distribution.simsorted
        
    @property
    def xsim(self):
        if self._u == 0:
            return self.x
        return self.distribution.mean
        
    @property
    def usim(self):
        if self._u == 0:
            return 0
        return self.distribution.stdev
        
    @property
    def cisim(self):
        if self._u == 0:
            return (self.x,self.x)

        if self._cimethod == 'shortest':
            return self.distribution.ci(self.p)
        else:
            return self.distribution.cisym(self.p)
              
    @property
    def Usim(self):
        if self._u == 0:
            return (0,0)
        x = self.distribution.mean
        
        if self._cimethod == 'shortest':
            ci = self.distribution.ci(self.p)
        else:
            ci = self.distribution.cisym(self.p)
            
        return (ci[1]-x,x-ci[0])
        
    @property
    def ksim(self):
        """
        read-only

        Returns ``0.5*(gummy.Usim[0] + gummy.Usim[1])/gummy.usim``
        """
        return 0.5*(self.Usim[0] + self.Usim[1])/self.usim
        
    @property
    def independent(self):
        """
        `bool`, read-only

        Returns `False` if the owning gummy was created from a operation involving
        other gummys and `True` otherwise.
        """
        if self._u == 0:
            return False
        return self._dist.isindependent
        
    @property
    def cimethod(self):
        """
        `str` in {'shortest', 'symmetric'}), default is 'shortest'
        
        Get or set the method for calculating the confidence interval from 
        Monte-Carlo data.  If this property is set at the class level, it will
        change the default `cimethod` value for new gummys but will no affect
        gummys that have already been created.
        
        Can be set either to the string 'shortest' or the string 'symmetric'.
        This property gets or sets the method for calculating confidence
        intervals from Monte-Carlo data.  
        
        If it is set to 'shortest', the confidence interval will be taken to be 
        the shortest interval that includes the desired fraction of the probability 
        distribution.  
        
        If it is set to 'symmetric', then the confidence interval will be set so 
        that, for n Monte-Carlo samples and a coverage probability of `p`, then
        `n`*(1-`p`)/2 samples lie below the lower limit of the confidence interval
        and the same number of samples lie above the upper limit of the confidence 
        interval.
        """
        
        return self._cimethod
    @cimethod.setter
    def cimethod(self,value):
        value = value.lower().strip()
        if value not in ['shortest','symmetric']:
            raise ValueError('cimethod ' + str(value) + ' is not recognized')
        self._cimethod = value
        
    @property
    def p(self):
        return 0.68268949213708585
    @p.setter
    def p(self,v):
        pass
        
    @staticmethod
    def set_seed(seed):
        """
        Sets the seed for the numpy.random.RandomState object shared by all 
        `Distribution` instances.
        """
        
        Distribution.set_seed(seed)
        
    @staticmethod
    def _copy(s,r,formatting=True,tofloat=False):
        # copies attributes of s to r, called from ummy.copy()
        super(nummy,nummy)._copy(s,r,formatting=formatting,tofloat=tofloat)
        if isinstance(s,nummy):
            r._dist = s._dist
            if formatting:
                r.name = s.name
                r._cimethod = s._cimethod
            else:
                r.name = None
        else:
            r.name = None
            if s._u == 0:
                r._dist = r._x
                return
            dof = s._dof
            if isinf(dof):
                r._dist = NormalDist(s._x,s._u)
            else:
                if r._bayesian:
                    r._dist = TDist(s._x,s._u*(dof-2)/dof,dof)
                    r._dof = float('inf')
                else:
                    r._dist = TDist(s._x,s._u,dof)
        
    @classmethod
    def _apply(cls,function,derivative,*args,fxdx=None):
        # called from ummpy.apply()
        if not nummy._mcpropagate:
            return super(nummy,cls)._apply(function,derivative,*args,fxdx=fxdx)
            
        a = list(args)
        for i,e in enumerate(args):
            if isinstance(e,nummy):
                a[i] = e._dist
            elif isinstance(e,ummy):
                return super(nummy,cls)._apply(function,derivative,*args,fxdx=fxdx)
        r = super(nummy,cls)._apply(function,derivative,*args,fxdx=fxdx)
        if isinstance(r,nummy):
            r._dist = Distribution.apply(function,*a)
        return r
        
    @classmethod
    def _napply(cls,function,*args,fxx=None):
        # called from ummpy.napply()
        if not nummy._mcpropagate:
            return super(nummy,cls)._napply(function,*args,fxx=fxx)
            
        a = list(args)
        for i,e in enumerate(args):
            if isinstance(e,nummy):
                a[i] = e._dist
            elif isinstance(e,ummy):
                return super(nummy,cls)._napply(function,*args,fxx=fxx)
        r = super(nummy,cls)._napply(function,*args,fxx=fxx)
        if isinstance(r,nummy):
            r._dist = Distribution.apply(function,*a)
        return r
        
    @staticmethod
    def _set_correlation_matrix(gummys, matrix):
        super(nummy,nummy)._set_correlation_matrix(gummys,matrix)
        cov = nummy.covariance_matrix(gummys)
        mean = [g.x for g in gummys]
        if isfinite(gummys[0].dof) and all([g.dof==gummys[0].dof for g in gummys]):
            mvdist = MultiTDist(mean,cov,gummys[0].dof)
        else:
            mvdist = MultiNormalDist(mean,cov)
        for i,g in enumerate(gummys):
            g._dist = mvdist[i]
            
    @staticmethod
    def _set_covariance_matrix(gummys, matrix):
        super(nummy,nummy)._set_covariance_matrix(gummys,matrix)
        mean = [g.x for g in gummys]
        if isfinite(gummys[0].dof) and all([g.dof==gummys[0].dof for g in gummys]):
            mvdist = MultiTDist(mean,matrix,gummys[0].dof)
        else:
            mvdist = MultiNormalDist(mean,matrix)
        for i,g in enumerate(gummys):
            g._dist = mvdist[i]
        
    @classmethod
    def create(cls,x,u=0,dof=float('inf'),name=None,correlation_matrix=None,
               covariance_matrix=None):
        if name is None:
            name = [None]*len(x)
                   
        if isinstance(x,MultivariateDistribution):

            # Do this the hard way because we want to raise an exception if
            # a MultivariateElement is passed the the nummy initializer in 
            # case it is tried outside create and the .ref is not set up
            # correctly.
            u = x.u()
            nd = len(u)
            if hasattr(x,'dof'):
                if nummy._bayesian:
                    dof = [float('inf')]*nd
                    for i,uu in u:
                        u[i] *= x.dof[i]/(x.dof[i]-2)
                else:
                    dof = x.dof
            else:
                dof = [float('inf')]*nd
            
            ret = super(nummy,cls).create(x.x(),u=u,dof=dof,covariance_matrix=x.cov)
                   
            for i,r in enumerate(ret):
                r._dist = x[i]
                r.name = name[i]
                
            return ret
                
        if any([isinstance(v,Distribution) for v in x]):
            if correlation_matrix is not None or covariance_matrix is not None:
                raise TypeError('Distribtuion instances may not be used in x if a correlation_matrix nor a covariance_matrix is defined')
            
        if dof is None:
            dof = [float('inf')]*len(x)
        if u is None:
            u = [0]*len(x)
        d = [None]*len(x)
        for i,v in enumerate(x):
            if isinstance(v,Distribution):
                d[i] = v
                x[i] = v.x()
                u[i] = v.u()
                if hasattr(v,'dof'):
                    dof[i] = v.dof
                    if nummy._bayesian:
                        u[i] = u[i]*v.dof/(v.dof-2)
                        dof[i] = float('inf')
            
        ret = super(nummy,cls).create(x,u,dof,correlation_matrix,covariance_matrix)
        for i,r in enumerate(ret):
            r.name = name[i]
            if d[i] is None:
                if u[i] == 0:
                    r._dist = x[i]
                elif isinf(dof[i]):
                    r._dist = NormalDist(x[i],u[i])
                else:
                    if nummy._bayesian:
                        r._dist = TDist(x,u[i]*(dof[i]-2)/dof[i],dof[i])
                        r._dof = float('inf')
                    else:
                        r._dist = TDist(x,u[i],dof[i])
            else:
                r._dist = d[i]
                    
        return ret

    def hist(self,xlabel=None,title=None,hold=False,**kwds):
        if xlabel is None and self.name is not None:
            xlabel = str(self.name)
            if len(xlabel) == 1:
                xlabel = '$ ' + xlabel + ' $'
        self.distribution.hist(xlabel=xlabel,title=title,hold=hold,**kwds)
        
    def covariance_sim(self,g):
        """
        Returns the covariance, calculated from Monte-Carlo data, between the 
        owning gummy and the gummy `g.`
        
        See the method `gummy.covariance(g)` for the corresponding result based
        on first order error propagation.
        """
        if self._u == 0 or g._u == 0:
            return 0
        return self._dist.covsim(g._dist)
        
    def correlation_sim(self,g):
        """
        Returns the correlation coefficient, calculated from Monte-Carlo data, 
        between the owning gummy and the gummy `g`.
        
        See the method `gummy.correlation(g)` for the corresponding result based
        on first order error propagation.
        """
        if self._u == 0 or g._u == 0:
            return 0
        return self.covariance_sim(g)/(self.usim*g.usim)
        
    @staticmethod
    def covariance_matrix_sim(gummys):
        """
        The staticmethod takes a list of gummys an returns the variance-covariance
        matrix calculated from Monte-Carlo data.  The return value is numpy
        ndarray.
        
        See the method gummy.covariance_matrix(gummys) for the corresponding
        result based on first order error propagation.
        """
        d = [v._dist for v in gummys]
        return Distribution.covsim_matrix(*d)
        
    @staticmethod
    def correlation_matrix_sim(gummys):
        """
        The staticmethod takes a list of gummys an returns the correlation
        matrix calculated from Monte-Carlo data.  The return value is numpy 
        ndarray.
        
        See the method `gummy.correlation_matrix(gummys)` for the corresponding
        result based on first order error propagation.
        """
        cov = nummy.covariance_matrix_sim(gummys)
        return [[cov[i][j]/sqrt(cov[i][i]*cov[j][j]) 
                for i in range(len(gummys))] for j in range(len(gummys))]
                    
    @staticmethod
    def covplot(x,y,title=None,xlabel=None,ylabel=None,hold=False,**kwds):
        if xlabel is None and x.name is not None:
            xlabel = str(x.name)
            if len(xlabel) == 1:
                xlabel = '$ ' + xlabel + ' $'
        if ylabel is None and y.name is not None:
            ylabel = str(y.name)
            if len(ylabel) == 1:
                ylabel = '$ ' + ylabel + ' $'

        Distribution.covplot(x.distribution,y.distribution,xlabel=xlabel,ylabel=ylabel,
                             title=title,hold=hold,**kwds)
        
    def ufrom(self,x,sim=False):
        """
        Gets the standard uncertainty contributed from particular gummys
        or utypes if all other free variables are held fixed.

        Parameters
        ---------_
        x:  `gummy`, `str`, or array_like
            A gummy, a string referencing a utype or a list containing
            gummys and strings.

        Returns
        -------
        `float`

        Example
        -------
        >>>  a = gummy(1.2,0.2,utype='A')
        >>>  b = gummy(3.2,0.5,utype='A')
        >>>  c = gummy(0.9,0.2,utype='B')
        >>>  d = a + b + c
        >>>  d.ufrom('A')
        0.53851648071345048
        """
        if not sim:
            return super().ufrom(x)
        
        x = [g for g in ummy._toummylist(x) if self.correlation(g) != 0]
            
        v = nummy.correlation_matrix_sim(x)
            
        b = [self.correlation(z) for z in x]
        s = np.linalg.lstsq(v,b)[0]
        u = 0
        
        d = [i*self.usim/j.usim for i,j in zip(s,x)]
        for i in range(len(x)):
            for j in range(len(x)):
                u += d[i]*d[j]*x[i].correlation(x[j])*x[i].usim*x[j].usim
                
        return sqrt(u)
        
    def _add(self,b):
        return _bop(super()._add(b),np.add,self,b)
        
    def _radd(self,b):
        return _rbop(super()._radd(b),np.add,self,b)
        
    def _sub(self,b):
        return _bop(super()._sub(b),np.subtract,self,b)
        
    def _rsub(self,b):
        return _rbop(super()._rsub(b),np.subtract,self,b)
        
    def _mul(self,b):
        return _bop(super()._mul(b),np.multiply,self,b)
        
    def _rmul(self,b):
        return _rbop(super()._rmul(b),np.multiply,self,b)
        
    def _truediv(self,b):
        return _bop(super()._truediv(b),np.divide,self,b)
        
    def _rtruediv(self,b):
        return _rbop(super()._rtruediv(b),np.divide,self,b)
        
    def _pow(self,b):
        return _bop(super()._pow(b),np.power,self,b)
        
    def _rpow(self,b):
        return _rbop(super()._rpow(b),np.power,self,b)
    
    def _nprnd(self,f):
        ret = super()._nprnd(f)
        if nummy._mcpropagate:
            ret._dist = Distribution.apply(f,self._dist)
        return ret
        
    def _mod(self,b):
        return _bop(super()._mod(b),np.mod,self,b)
    
    def _rmod(self,b):
        return _rbop(super()._rmod(b),np.mod,self,b)
        
    def __neg__(self):
        return _uop(super().__neg__(),np.negative,self)
        
    def __pos__(self):
        return _uop(super().__pos__(),np.positive,self)
        
    def __abs__(self):
        return _uop(super().__abs__(),np.absolute,self)